#ifndef INCLUDE__InformationManager_hxx
#define INCLUDE__InformationManager_hxx

#include <log4cxx/logger.h>
#include <boost/signals2.hpp>
#include <functional>
#include <string>
#include <map>

#include "disassembler/Disassembler.hxx"

class Function;
class BasicBlock;
class Comment;

class QString;

class RenameFunctionEvent;

class InformationManager {
public:
	~InformationManager();

	void reset(const std::string& filename);
	void save(const QString& filename);

	void signal_new_function(Function* f);
	void signal_new_dyn_symbol(const std::string& f)
		{ new_dyn_symbol_signal(f);	}

	boost::signals2::connection
	connect_new_function_signal(std::function<void(Function*)> f)
		{ return new_function_signal.connect(f); }

	boost::signals2::connection
	connect_new_dyn_symbol_signal(std::function<void(const std::string& name)> f)
		{ return new_dyn_symbol_signal.connect(f); }

	boost::signals2::connection
	connect_reset_signal(std::function<void ()> f)
		{ return reset_signal.connect(f); }

	boost::signals2::connection
	connect_rename_function_signal(std::function<void (RenameFunctionEvent*)> f)
		{ return rename_function_signal.connect(f); }

	Disassembler* getDisassembler()
		{ return disassembler.get(); }

	void dispatch(RenameFunctionEvent* event)
		{ rename_function_signal(event); }

	Function* getFunction(uint64_t address);
	BasicBlock* getBasicBlock(uint64_t address);
	bool hasFunctions() const {return functions.size() != 0;}

	/* Protocoll:
	 *
	 * Users may allocate new Data containers with the new*()
	 * functions. Once they have populated the information they hand
	 * over the object to the information manager using the finish*()
	 * functions.
	 *
	 * if new*() returns NULL there already exists a function at the
	 * specified address. Users may then get the old object if they
	 * wish or (more likely) skip creating it. Uniqueness of the
	 * object is only guaranteed as compared to the finish()ed
	 * objects.
	 *
	 * Users are responsible for destroying functions iff they do not
	 * finish them using the delete*() functions. Once the objects are
	 * finished, the information manager is responsible for cleaning
	 * up the memory. If delete*() is called on a finished object, bad
	 * thingsmay happen.
	 */
	Function* newFunction(uint64_t address);
	BasicBlock* newBasicBlock(uint64_t address);
	Comment* newGlobalComment(uint64_t address);
	Comment* newLocalComment(uint64_t address, Function* f);
	void finishFunction(Function* f);
	void finishBasicBlock(BasicBlock* b);
	void finnishComment(Comment* c);
	void deleteFunction(Function* f);
	void deleteBasicBlock(BasicBlock* b);
	void deleteComment(Comment* c);
private:
	boost::signals2::signal<void ()> reset_signal;
	boost::signals2::signal<void (Function*)> new_function_signal;
	boost::signals2::signal<void (const std::string& name)> new_dyn_symbol_signal;
	boost::signals2::signal<void (RenameFunctionEvent*)> rename_function_signal;
	std::unique_ptr<Disassembler> disassembler;
	std::map<uint64_t, Function*> functions;
	std::map<uint64_t, BasicBlock*> blocks;
	std::string filename;

	log4cxx::LoggerPtr logger;
};

#endif /* INCLUDE__InformationManager_hxx */
